%% demo code multiclass segmentation (authored by Zhenzhang Ye & Tao Wu)


%% settings
close all
clear all

show_img_at_it = true;
scale_img_size = .25;

gap_tol = 1e-4;
max_iters = 500;

% method = 'pdhg';
% method = 'drs';
% method = 'admm';
method = 'all';


%% read image

img = imread('butterfly.png');
img = double(imresize(img,scale_img_size))/255.0;
r = img(:,:,1);
g = img(:,:,2);
b = img(:,:,3);
[ny,nx,c] = size(img);
n = ny*nx;

k = 4;
[~,C] = kmeans(reshape(img, [n,3]),k);

%% compute data term f
f = zeros(ny,nx,k);
for i = 1:k
    f(:,:,i) = sqrt(sum((img-repmat(reshape(C(i,:), 1,1,3),ny,nx)).^2,3));
end
f = reshape(f,n,k);

%% construct the linear operater K
e = ones(n, 1);
Dx_tilde = spdiags([e -e], -1:0, nx, nx); 
Dx_tilde(:, end) = 0;
e = ones(nx, 1);
Dy_tilde = spdiags([-e e], 0:1, ny, ny);
Dy_tilde(end, end) = 0;
Dx = kron(Dx_tilde', speye(ny));
Dy = kron(speye(nx), Dy_tilde);
alpha = .1;
D = [Dx; Dy] * alpha;
normD = sqrt(8) * alpha;
% normD = normest(D,1e-6);


%% initialize general parameters
% gap_tol = 1e-6;
% max_iters = 1000;
% 
% method = 'pdhg';

% return


%% PDHG
if strcmp(method, 'pdhg') || strcmp(method, 'all')
    
u = zeros(n,k);
u_tilde = zeros(n,k);
p = zeros(2*n, k);
s = 1.001*normD;
t = 1.001*normD;

if(show_img_at_it)
    figure(1)
    hold on;
end

primal_ener_pdhg = nan(1,max_iters);
dual_ener_pdhg = nan(1,max_iters);

tic;
for it=1:max_iters
    
    % store previous iterate
    u_prev = u;
    
    % primal update
    u_tilde = u - 1/s * (D' * p + f);
    u = projSimplex(u_tilde);
    
    % dual update
    p = min(max( p +  D * (2*u - u_prev)/t, -1), 1);
    
    % compute energies, duality gap
    primal_ener_pdhg(it) = (sum(reshape(u .* f, [],1)) + sum(reshape(abs(D * u), [],1)))/n;
    dual_ener_pdhg(it) = sum(max( -D'*p-f, [],2))/n;
    gap = primal_ener_pdhg(it) + dual_ener_pdhg(it);
%     return
    
    % plot
    if(show_img_at_it)
        [~, classes] = max(u, [], 2);
        im_result = reshape(C(classes, :), [ny nx 3]);

        figure(100)
        if it==1
            subplot(121)
            imagesc(img);
            title('image')
            axis equal off
        end

        subplot(122)
        imagesc(im_result);
        title(['PDHG at iter ',num2str(it)])
        axis equal off
        drawnow;
    end
    
    % stopping rule
    if gap<gap_tol
        fprintf('iter %d: gap= %.3e, primal-ener= %.3e\n', it, gap, primal_ener_pdhg(it));
        fprintf('PDHG terminated!\n');
        
        break
        
    elseif mod(it,10)==0 || it==1
        fprintf('iter %d: gap= %.3e, primal-ener= %.3e\n', it, gap, primal_ener_pdhg(it));
        
    end
    
end
toc

u_pdhg = u;

end

% return


%% DRS
if strcmp(method, 'drs') || strcmp(method, 'all')

u  = zeros(n,k);
v = zeros(size(u));
tau = 1.;
inner_its = 10;     % num of inner iter for "inner_dual"

primal_ener_drs = nan(1,max_iters);
dual_ener_drs = nan(1,max_iters);

tic;
for it = 1:max_iters
    
    u = projSimplex(v - tau*f);
    v = v - u + inner_dual(2*u-v, D, tau, normD, inner_its);
    
    % generate "multiplier"
    p = (u - v) / tau;
    
    primal_ener_drs(it) = (sum(reshape(u .* f, [],1)) + sum(reshape(abs(D * u), [],1)))/n;
    dual_ener_drs(it) = sum(max( -p-f, [],2))/n;
    gap = primal_ener_drs(it) + dual_ener_drs(it);
    
    % plot
    if(show_img_at_it)
        [~, classes] = max(u, [], 2);
        im_result = reshape(C(classes, :), [ny nx 3]);

        figure(100)
        if it==1
            subplot(121)
            imagesc(img);
            title('image')
            axis equal off
        end
        subplot(122)
        imagesc(im_result);
        title(['DRS at iter ',num2str(it)])
        axis equal off
        drawnow;
    end
    
    % stopping rule
    if gap<gap_tol
        fprintf('iter %d: gap= %.3e, primal-ener= %.3e\n', it, gap, primal_ener_drs(it));
        fprintf('DRS terminated!\n');
        break
        
    elseif mod(it,10)==0 || it==1
        fprintf('iter %d: gap= %.3e, primal-ener= %.3e\n', it, gap, primal_ener_drs(it));
        inner_its = min(round(inner_its * 1.2), 1e3);   % increase num of inner iter
        
    end
    
end
toc

u_drs = u;

end

% return


%% ADMM 
if strcmp(method, 'admm') || strcmp(method, 'all')

% u = zeros(n,k);
% v1 = zeros(n,k);
% v2 = zeros(2*n,k);
% p1 = zeros(n,k);
% p2 = zeros(2*n,k);
% rho = 1e2;
% M = eye(size(D,2))+D'*D;

u = zeros(n,k);
v = zeros(n,k);
p = zeros(n,k);
tau = 1.;
inner_its = 10;     % num of inner iter for "inner_dual"

primal_ener_admm = nan(1,max_iters);
dual_ener_admm = nan(1,max_iters);

tic;
for it=1:max_iters
    
    u = projSimplex(v - (p+f)/tau);
    
    v = inner_dual(u + p/tau, D, 1/tau, normD, inner_its);
    
    p = p + tau * (u - v);
    
    primal_ener_admm(it) = (sum(reshape(u .* f, [], 1)) + sum(reshape(abs(D * u), [], 1)))/n;
    dual_ener_admm(it) = sum(max(-p-f,[],2))/n;
    gap = (primal_ener_admm(it)+dual_ener_admm(it));
    
    % plot
    if(show_img_at_it)
        [~, classes] = max(u, [], 2);
        im_result = reshape(C(classes, :), [ny nx 3]);

        figure(100)
        if it==1
            subplot(121)
            imagesc(img);
            title('image')
            axis equal off
        end
        subplot(122)
        imagesc(im_result);
        title(['ADMM at iter ',num2str(it)])
        axis equal off
        drawnow;
    end
    
    % stopping rule
    if gap<gap_tol
        fprintf('iter %d: gap= %.3e, primal-ener= %.3e\n', it, gap, primal_ener_admm(it));
        fprintf('ADMM terminated!\n');
        break
        
    elseif mod(it,10)==0 || it==1
        fprintf('iter %d: gap= %.3e, primal-ener= %.3e\n', it, gap, primal_ener_admm(it));
        inner_its = min(round(inner_its * 1.2), 1e3);   % increase num of inner iter
        
    end
    
end
toc

u_admm = u;

end

% return


%% visualize results

visualize_result
